#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestCopyWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('copy_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',schema=('ID*:int64','symbol:string','time:stamp'))
        src.add_connector('fs',
                          conn_name='fsin', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='copy_window.csv',
                          transactional=True,
                          blocksize='1',
                          dateformat='%Y-%m-%d %H:%M:%S')

        cp_win = esp.CopyWindow(name='cp_win')
        cp_win.set_retention('bytime_sliding','300 seconds',field='time')
        cp_win.add_connector('fs',
                             conn_name='fsout', conn_type='subscribe',
                             type='sub',
                             fstype='csv',
                             fsname='output',
                             snapshot=True,
                             dateformat='%Y-%m-%d %H:%M:%S')

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(cp_win)

        src.add_target(cp_win)

        self._test_model_file('copy_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import Window, attribute
from .utils import get_args


class CounterWindow(Window):
    '''
    Counter window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single
        update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.
    count_interval : string, optional
        Specifies the interval period at which the Counter window
        generates events.
    clear_interval : string, optional
        Specifies the interval of inactivity after which the values in
        the Counter window are cleared.

    Returns
    -------
    :class:`CounterWindow`

    '''

    window_type = 'counter'

    count_interval = attribute('count-interval', dtype='string')
    clear_interval = attribute('clear-interval', dtype='string')

    def __init__(self, name=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_max_string=None, index_type=None,
                 pubsub_index_type=None, count_interval=None,
                 clear_interval=None):
        Window.__init__(self, **get_args(locals()))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Sniffer Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class SnifferPublisher(Connector):
    '''
    Publish local area network packet events

    Parameters
    ----------
    interface : string
        Specifies the name of the network interface on the local
        machine from which to capture packets.
    protocol : string
        Specifies the port number associated with the protocol type
        of packets to be captured. You can specify this as a
        comma-separated list of port numbers.
    packetfields : string
        Specifies the packet fields to be extracted from a captured
        packet and included in the published event.
    transactional : string, optional
        Sets the event block type to transactional. The default
        value is normal.
    blocksize : int, optional
        Specifies the number of events to include in a published
        event block. The default value is 1.
    addtimestamp : boolean, optional
        Specifies to append an ESP_TIMESTAMP field to each published event.
    configfilesection : string, optional
        Specifies the name of the section in the configuration file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    vendorid : string, optional
        Specifies the vendor-Id field to match when capturing the
        Attribute-Specific field in a Vendor-Specific attribute in a
        Radius Accounting-Request packet.
    vendortype : string, optional
        Specifies the vendor-Type field to match when capturing the
        Attribute-Specific field in a Vendor-Specific attribute in a
        Radius Accounting-Request packet.
    indexfieldname : string, optional
        Specifies the name to use instead of index for the index:int64
        field in the Source window schema.
    publishwithupsert : boolean, optional
        Specifies to build events with opcode = Upsert instead of Insert.
    pcapfilter : string, optional
        Specifies a filter expression as defined in the pcap
        documentation. Passed to the pcap driver to filter packets
        received by the connector.
    httpports : string, optional
        Specifies a comma-separated list of destination ports. All
        sniffed packets that contain a specified port are parsed for
        HTTP GET parameters. The default value is 80.
    ignorenopayloadpackets : boolean, optional
        Specifies whether to ignore packets with no payload, as
        calculated by subtracting the TCP or UDP header size from the
        packet size. The default value is FALSE.
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`SnifferPublisher`

    '''
    connector_key = dict(cls='sniffer', type='publish')

    property_defs = dict(
        interface=prop('interface', dtype='string', required=True),
        protocol=prop('interface', dtype='string', required=True),
        packetfields=prop('packetfields', dtype='string', required=True),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        addtimestamp=prop('addtimestamp', dtype='boolean'),
        configfilesection=prop('configfilesection', dtype='string'),
        vendorid=prop('vendorid', dtype='string'),
        vendortype=prop('vendortype', dtype='string'),
        indexfieldname=prop('indexfieldname', dtype='string'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        pcapfilter=prop('pcapfilter', dtype='string'),
        httpports=prop('httpports', dtype='string'),
        ignorenopayloadpackets=prop('ignorenopayloadpackets', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, interface=None, protocol=None, packetfields=None, name=None,
                 is_active=None, transactional=None, blocksize=None,
                 addtimestamp=None, configfilesection=None, vendorid=None,
                 vendortype=None, indexfieldname=None, publishwithupsert=None,
                 pcapfilter=None, httpports=None, ignorenopayloadpackets=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'sniffer', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['interface',
                                                   'protocol',
                                                   'packetfields'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import base64
import copy
import numpy as np
import os
import pandas as pd
import six
import unittest
from PIL import Image
from . import utils as tm
from ..transformers import bgr2rgb, rgb2bgr, bytes2image

DATA_DIR = tm.get_data_dir()


class TestTransformers(tm.TestCase):

    def test_rgb2bgr(self):
        # Images
        rgb1 = Image.fromarray(np.asarray(Image.open(os.path.join(DATA_DIR, 'scooter.jpg'))))
        rgb2 = bgr2rgb(rgb2bgr(Image.open(os.path.join(DATA_DIR, 'scooter.jpg'))))

        bgr1 = rgb2bgr(Image.open(os.path.join(DATA_DIR, 'scooter.jpg')))
        bgr2 = rgb2bgr(bgr2rgb(rgb2bgr(Image.open(os.path.join(DATA_DIR, 'scooter.jpg')))))

        self.assertEqual(list(rgb1.getdata()), list(rgb2.getdata()))
        self.assertEqual(list(bgr1.getdata()), list(bgr2.getdata()))
        self.assertNotEqual(list(bgr1.getdata()), list(rgb1.getdata()))

        # DataFrame
        df_rgb1 = pd.DataFrame([['a', rgb1], ['b', rgb2]], columns=['Name', 'Image'])
        df_rgb2 = bgr2rgb(rgb2bgr(df_rgb1))

        self.assertEqual(list(df_rgb1.columns), ['Name', 'Image'])
        self.assertEqual(list(df_rgb2.columns), ['Name', 'Image'])
        self.assertEqual(list(df_rgb1['Image'][0].getdata()),
                         list(df_rgb2['Image'][0].getdata()))
        self.assertEqual(list(df_rgb1['Image'][1].getdata()),
                         list(df_rgb2['Image'][1].getdata()))
        
        # DataFrame with columns
        df_rgb1 = pd.DataFrame([['a', rgb1], ['b', rgb2]], columns=['Name', 'Image'])
        df_rgb2 = bgr2rgb(rgb2bgr(df_rgb1, columns='Image'), columns='Image')

        self.assertEqual(list(df_rgb1.columns), ['Name', 'Image'])
        self.assertEqual(list(df_rgb2.columns), ['Name', 'Image'])
        self.assertEqual(list(df_rgb1['Image'][0].getdata()),
                         list(df_rgb2['Image'][0].getdata()))
        self.assertEqual(list(df_rgb1['Image'][1].getdata()),
                         list(df_rgb2['Image'][1].getdata()))
        
        # Unknown data
        self.assertEqual(bgr2rgb(10), 10)
        self.assertEqual(rgb2bgr(10), 10)

    def test_bytes2image(self):
        # Image
        jpg = open(os.path.join(DATA_DIR, 'scooter.jpg'), 'rb').read()
        img = bytes2image(jpg)
        self.assertTrue(Image.isImageType(img))

        # DataFrame
        df_jpg = pd.DataFrame([['a', jpg]], columns=['Name', 'Image'])
        df_jpg = bytes2image(df_jpg)
        self.assertEqual(list(df_jpg.columns), ['Name', 'Image'])
        self.assertTrue(Image.isImageType(df_jpg['Image'][0]))

        # DataFrame with columns
        df_jpg = pd.DataFrame([['a', jpg]], columns=['Name', 'Image'])
        df_jpg = bytes2image(df_jpg, columns='Image')
        self.assertEqual(list(df_jpg.columns), ['Name', 'Image'])
        self.assertTrue(Image.isImageType(df_jpg['Image'][0]))

        # Unknown data
        self.assertEqual(bytes2image(10), 10)


if __name__ == '__main__':
    tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Connectors '''

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import Connector
from .bacnet import BacnetPublisher
from .adapter import AdapterPublisher, AdapterSubscriber
from .database import DatabasePublisher, DatabaseSubscriber
from .fs import FilePublisher, FileSubscriber, SocketPublisher, SocketSubscriber
from .kafka import KafkaSubscriber, KafkaPublisher
from .mqtt import MQTTSubscriber, MQTTPublisher
from .modbus import ModbusSubscriber, ModbusPublisher
from .nurego import NuregoSubscriber
from .opcua import OPCUASubscriber, OPCUAPublisher
from .pi import PISubscriber, PIPublisher
from .project import ProjectPublisher
from .pylon import PylonPublisher
from .rabbitmq import RabbitMQSubscriber, RabbitMQPublisher
from .smtp import SMTPSubscriber
from .sniffer import SnifferPublisher
from .solace import SolaceSubscriber, SolacePublisher
from .teradata import TeradataSubscriber, TeradataListenerSubscriber
from .tervela import TervelaSubscriber, TervelaPublisher
from .tibco import TibcoSubscriber, TibcoPublisher
from .timer import TimerPublisher
from .url import URLPublisher
from .uvc import UVCPublisher
from .websocket import WebSocketPublisher
from .websphere import WebSphereMQSubscriber, WebSphereMQPublisher



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Connector Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import numbers
import re
import six
from ..base import ESPObject
from ..utils import xml
from ..utils.data import gen_name


RegexType = type(re.compile(r''))


def listify(obj):
    if obj is None:
        return
    if isinstance(obj, (tuple, set, list)):
        return list(obj)
    return [obj]


def map_properties(cls, props, required=None, delete=None):
    '''
    Remap property names

    Parameters
    ----------
    cls : Connector-subclass
        The Connector class which contains the property definitions
    props : dict
        The dictionary of properties
    required : string or list-of-strings, optional
        The list of required keys.  These will be returned in a tuple
        in addition to the remaining dictionary of properties.
    delete : string or list-of-strings, optional
        List of keys to remove

    Returns
    -------
    ((required-params), {properties})
        If ``required`` contains names
    {properties}
        If ``required` is empty

    '''
    delete = listify(delete) + ['self']
    required = listify(required)

    names = {}
    names.update({v.name: k for k, v in six.iteritems(cls.property_defs)})
    names.update({k: k for k, v in six.iteritems(cls.property_defs)})

    out = {}
    for key, value in six.iteritems(props):
        if delete and key in delete:
            continue
        try:
            out[names[key]] = value
        except KeyError:
            raise KeyError('%s is not a valid parameter for %s' % (key, cls))

    if required:
        req_out = []
        for item in required:
            req_out.append(out.pop(item, None))
        return tuple(req_out), out

    return out


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        for subcls in get_subclasses(subclass):
            yield subcls
        yield subclass


def get_connector_class(elem_or_class, type=None, properties=None):
    '''
    Get a connector class that matches the current element

    Parameters
    ----------
    elem_or_class : string or Element
        The name of the connector class, or an XML definition
    type : string, optional
        The type of the connector.  Ignored if ``elem_or_class`` is an Element.
    properties : dict, optional
        The properties for the connector.  Ignored if ``elem_or_class` is
        an Element.

    Returns
    -------
    :class:`Connector` or subclass of :class:`Connector`

    '''
    if isinstance(elem_or_class, six.string_types):
        if elem_or_class.startswith('<'):
            elem_or_class = xml.ensure_element(elem_or_class)

    if isinstance(elem_or_class, six.string_types):
        cls = elem_or_class

        if type is None:
            type = 'subscribe'

        if properties is None:
            properties = {}

    else:
        elem = elem_or_class

        cls = elem.attrib['class']

        if 'type' in elem.attrib:
            type = elem.attrib['type']
        else:
            type = elem.find('./properties/property[@name="type"]').text

        properties = {}
        for item in elem.findall('./properties/property'):
            properties[item.attrib['name']] = item.text

    if type.startswith('p'):
        type = 'publish'
    else:
        type = 'subscribe'

    out = []
    for item in get_subclasses(Connector):
        if item.connector_key['cls'] == cls and item.connector_key['type'] == type:
            out.append(item)

    if not out:
        return Connector

    if len(out) == 1:
        return out[0]

    # Check extra matching properties
    matches = True
    for item in reversed(sorted(out, key=lambda x: len(x.connector_key))):
        matches = True
        for key, value in six.iteritems(item.connector_key):
            if key == 'cls' or key == 'type':
                continue
            elif isinstance(value, RegexType):
                eprop = properties.get(key)
                if eprop is None or not value.match(eprop):
                    matches = False
                    break
            else:
                eprop = properties.get(key)
                if eprop is None or value != eprop:
                    matches = False
                    break
        if matches:
            break

    if matches:
        return item

    return Connector


def prop(name, dtype, required=False, valid_values=None, valid_expr=None, default=None):
    return ConnectorProperty(name, dtype, required=required,
                             valid_values=valid_values,
                             valid_expr=valid_expr,
                             default=None)


class ConnectorProperty(object):

    def __init__(self, name, dtype, required=False, valid_values=None,
                 valid_expr=None, default=None):
        self.name = name
        self.dtype = dtype
        self.required = required

        if valid_values is None:
            self._valid_values = None
        elif isinstance(valid_values, (six.string_types, RegexType)):
            self._valid_values = [valid_values]
        else:
            self._valid_values = list(valid_values)

        if valid_expr is None:
            self._valid_expr = None
        elif isinstance(valid_expr, (six.string_types, RegexType)):
            self._valid_expr = [valid_expr]
        else:
            self._valid_expr = list(valid_expr)

        if default is None:
            self.default = default
        else:
            self.default = self.validate_value(default)

    def validate_type(self, value):
        '''
        Verify that the given value is the correct type

        Parameters
        ----------
        value : any
            The property value

        Raises
        ------
        TypeError
            If the value is not the declared type

        '''
        unmatched_types = []

        dtypes = self.dtype
        if not isinstance(self.dtype, (list, tuple)):
            dtypes = [self.dtype]

        for dtype in dtypes:
            if dtype in ['int', int]:
                if not isinstance(value, numbers.Integral):
                    unmatched_types.append(dtype)
            elif dtype in ['float', 'double', float]:
                if not isinstance(value, (numbers.Integral, numbers.Real)):
                    unmatched_types.append(dtype)
            elif dtype in ['boolean', 'bool', bool]:
                if value is not True and value is not False:
                    unmatched_types.append(dtype)
            elif dtype in ['string', str]:
                if not isinstance(value, six.string_types):
                    unmatched_types.append(dtype)
            else:
                raise TypeError('Unknown data type: %s' % dtype)

        if len(unmatched_types) == len(dtype):
            raise TypeError('%s is not one of: %s' % ', '.join(unmatched_types))

    def validate_value(self, value):
        '''
        Verify that the value is valid

        Parameters
        ----------
        value : any
            The property value to test

        Returns
        -------
        bool

        '''
        if value is None:
            return not(self.required)

        # If it's an environment variable, always return true
        if isinstance(value, six.string_types):
            print(bool(re.search(r'@\w+@', value)))
        if isinstance(value, six.string_types) and bool(re.search('r@\w+@', value)):
            return True

        # Make sure value is the correct type
        self.validate_type(value)

        # Check specific values and regular expressions
        valid = True
        if self._valid_values:
            regexes = [x for x in self._valid_values if isinstance(x, RegexType)]
            values = [x for x in self._valid_values if not isinstance(x, RegexType)]
            if valid and values and value not in values:
                valid = False
            if valid and regexes:
                for item in regexes:
                    if not item.search(value):
                        valid = False
                        break

        # Check expressions and regular expressions
        if valid and self._valid_expr:
            for item in self._valid_expr:
                if isinstance(item, RegexType) and not re.search(item, value):
                    valid = False
                elif isinstance(item, six.string_types):
                    if not eval(item):
                        valid = False
                        break

        return valid


class Connector(collections.MutableMapping):
    '''
    Window connector

    Parameters
    ----------
    conncls : string
        The connecter class name
    type : string, optional
        The type of the connector
    name : string, optional
        The name of the connector
    properties : dict, optional
        Dictionary of connector properties

    '''
    connector_key = dict(cls='', type='')
    property_defs = dict()

    def __init__(self, conncls, type=None, name=None, is_active=None, properties=None):
        self.cls = conncls
        self.name = name or gen_name(prefix='c_')
        self.type = type
        self.is_active = is_active
        self.properties = {}
        self.set_properties(**properties)

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Returns
        -------
        :class:`Connector`

        '''
        return type(self).from_parameters(conncls=self.cls, type=self.type,
                                          name=self.name, is_active=self.is_active,
                                          properties=self.properties)

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo):
        return self.copy(deep=True)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        return cls(conncls, type=type, name=name, is_active=None,
                   properties=properties)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Construct connector from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            The XML definition
        session : requests.Session, optional
            The session object associated with the server

        Returns
        -------
        :class:`Connector`

        '''
        data = xml.ensure_element(data)

        conncls = data.attrib['class']
        name = data.attrib.get('name')
        type = data.attrib.get('type')
        is_active = data.attrib.get('active')

        properties = {}
        for item in data.findall('./properties/property'):
            properties[item.attrib['name']] = item.text

        return get_connector_class(data).from_parameters(conncls, type=type,
                                                         name=name,
                                                         is_active=is_active,
                                                         properties=properties)

    from_xml = from_element

    def to_element(self):
        '''
        Export connector definition to ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('connector', attrib={'class': self.cls,
                                                'name': self.name,
                                                'active': self.is_active,
                                                'type': self.type})

        sorted_items = sorted([(k, v)for k, v in six.iteritems(self.properties)])
        properties = collections.OrderedDict(sorted_items)

        # Add defaults
        for key, value in six.iteritems(type(self).property_defs):
            if value.default is not None and properties.get(key, None) is None:
                xml.add_properties(out, {key: value.default})

        # Always make sure configfilesection= is next
        if properties.get('configfilesection', None) is not None:
            xml.add_properties(out, dict(configfilesection=properties['configfilesection']))
            properties.pop('configfilesection')

        # Add the rest
        xml.add_properties(out, properties)

        return out

    def to_xml(self, pretty=False):
        '''
        Export connector definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the output include whitespace for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def set_properties(self, **kwargs):
        '''
        Set connector properties

        Parameters
        ----------
        **kwargs : keyword-parameters, optional
            Key / value pairs of properties

        '''
        for key, value in six.iteritems(kwargs):
            if key == 'type':
                if value.startswith('p'):
                    self.type = 'publish'
                else:
                    self.type = 'subscribe'
                continue
            if key == 'name':
                self.name = value
                continue
            self.properties[key] = value
        self.properties = {k: v for k, v in six.iteritems(self.properties)
                           if v is not None}

    def __getitem__(self, key):
        return self.properties[key]

    def __setitem__(self, key, value):
        self.properties[key] = value

    def __delitem__(self, key):
        del self.properties[key]

    def __iter__(self):
        return iter(self.properties)

    def __len__(self):
        return len(self.properties)

    def __str__(self):
        if self.type and self.name:
            return '%s(%s, name=%s, type=%s, properties=%s)' % \
                   (type(self).__name__, repr(self.cls), repr(self.name),
                    repr(self.type), repr(self.properties))
        if self.type:
            return '%s(%s, type=%s, properties=%s)' % \
                   (type(self).__name__, repr(self.cls),
                    repr(self.type), repr(self.properties))
        return '%s(%s, name=%s, properties=%s)' % \
               (type(self).__name__, repr(self.cls),
                repr(self.name), repr(self.properties))

    def __repr__(self):
        return str(self)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
from . import utils as tm


class TestAggregateWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('aggregate_window.xml')

    def test_field_expressions(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('tradeID*:string','security:string',
                                       'quantity:int32','price:double',
                                       'traderID:int64','dateD:date','timeD:stamp'))
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                           type='pub',
                           fstype='csv',
                           fsname='aggregate_window.csv',
                           transactional=True,
                           dateformat='%Y-%m-%d %H:%M:%S')

        aggregate_win = esp.AggregateWindow(name='aggregate',
                                            schema=('traderID*:int64',
                                                    'cat_security:string',
                                                    'cat_quantity:string',
                                                    'cat_price:string',
                                                    'cat_dateD:string',
                                                    'cat_timeD:string'))
        aggregate_win.add_field_expression('ESP_aCat(security,",")')
        aggregate_win.add_field_expression('ESP_aCat(quantity,",")')
        aggregate_win.add_field_expression('ESP_aCat(price,",")')
        aggregate_win.add_field_expression('ESP_aCat(dateD,",")')
        aggregate_win.add_field_expression('ESP_aCat(timeD,",")')
        aggregate_win.add_connector('fs',
                                   conn_name='fsout', conn_type='subscribe',
                                   type='sub',
                                   fstype='csv',
                                   fsname='result.out',
                                   snapshot=True,
                                   dateformat='%Y-%m-%d %H:%M:%S')

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(aggregate_win)

        src.add_target(aggregate_win)

        self._test_model_file('aggregate_window.xml', proj)

    @unittest.skip('Need way to load plugin')
    def test_plugin(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto', n_threads='1')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('ID*:int32', 'symbol:string',
                                       'quantity:int32', 'price:double'))
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='aggregate_with_plugin.csv',
                          transactional=True)

        aggregate_win = esp.AggregateWindow(name='aw',
                                            schema=('symbol*:string',
                                                    'totalQuant:int32',
                                                    'maxPrice:double'))
        aggregate_win.add_field_plugin(plugin='plugin',function='summationAggr',additive=True)
        aggregate_win.add_field_plugin(plugin='plugin',function='maximumAggr',additive_insert_only=True)
        aggregate_win.add_connector('fs',
                                   conn_name='sub', conn_type='subscribe',
                                   type='sub',
                                   fstype='csv',
                                   fsname='result.out',
                                   snapshot=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(aggregate_win)
        src.add_target(aggregate_win)

        self._test_model_file('aggregate_with_plugin.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import re
import six
from .base import Window, attribute
from .features import InitializeExpressionFeature
from .utils import get_args, ensure_element, connectors_to_end
from ..utils import xml


class FieldExpression(object):
    '''
    Join field expression

    Parameters
    ----------
    name : string
        Output field for the join
    expr : string
        Join expression
    type : string, optional
        Type of the output field

    Returns
    -------
    :class:`FieldExpression`

    '''

    def __init__(self, name, expr, type='double'):
        self.name = name
        self.type = type
        self.expr = expr

    def copy(self, deep=False):
        return type(self)(self.name, self.expr, type=self.type)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldExpression`

        '''
        data = ensure_element(data)
        return cls(data.attrib['name'], data.text, type=data.attrib['type'])

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('field-expr',
                            attrib=dict(name=self.name, type=self.type),
                            text_content=self.expr)

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class LeftExpression(object):
    '''
    Join left expression

    Parameters
    ----------
    names : string or list-of-strings, optional
        Specify fields from the left input to the Join window
        to use in output elements.

    Returns
    -------
    :class:`LeftExpression`

    '''

    def __init__(self, names='*'):
        if isinstance(names, six.string_types):
            names = re.split(r'\s*,\s*', names.strip())
        self.names = names or []

    def copy(self, deep=False):
        return type(self)(self.names)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldExpression`

        '''
        return cls(ensure_element(data).text)

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('left-expr', text_content=','.join(self.names))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class RightExpression(object):
    '''
    Join right expression

    Parameters
    ----------
    names : string or list-of-strings, optional
        Specify fields from the right input to the Join window
        to use in output elements.

    Returns
    -------
    :class:`RightExpression`

    '''

    def __init__(self, names='*'):
        if isinstance(names, six.string_types):
            names = re.split(r'\s*,\s*', names.strip())
        self.names = names or []

    def copy(self, deep=False):
        return type(self)(self.names)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldExpression`

        '''
        return cls(ensure_element(data).text)

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('right-expr', text_content=','.join(self.names))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class FieldPlugin(object):
    '''
    Join field plugin

    Parameters
    ----------
    name : string
        Name of the output field
    type : string
        Type of the output field
    plugin : string
        Full path to .so / .dll
    function : string
        Function to call in the library

    Returns
    -------
    :class:`FieldPlugin`

    '''

    def __init__(self, name, type, plugin, function):
        self.name = name
        self.type = type
        self.plugin = plugin
        self.function = function

    def copy(self, deep=False):
        return type(self)(self.name, self.type, self.plugin, self.function)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldPlugin`

        '''
        data = ensure_element(data)
        return cls(data.attrib['name'], data.attrib['type'],
                   data.attrib['plugin'], data.attrib['function'])

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('field-plug', attrib=dict(name=self.name,
                                                      type=self.type,
                                                      plugin=self.plugin,
                                                      function=self.function))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class FieldSelection(object):
    '''
    Join field selection

    Parameters
    ----------
    name : string
        Name of the field in the join schema
    source : string
        Field in the input schema

    Returns
    -------
    :class:`FieldSelection`

    '''

    def __init__(self, name, source):
        self.name = name
        self.source = source

    def copy(self, deep=False):
        return type(self)(self.name, self.source)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldSelection`

        '''
        return cls(data.attrib['name'], data.attrib['source'])

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('field-selection',
                            attrib=dict(name=self.name, source=self.source))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class LeftSelection(object):
    '''
    Join left selection

    Parameters
    ----------
    names : string or list-of-strings
        Specify fields from the left input to the Join window
        to use in output elements.
    exclude : string or list-of-strings, optional
        Specifies a comma-separated list of fields to use
        from the input window.

    Returns
    -------
    :class:`LeftSelection`

    '''

    def __init__(self, names, exclude=None):
        if isinstance(names, six.string_types):
            names = re.split(r'\s*,\s*', names.strip())
        if isinstance(exclude, six.string_types):
            exclude = re.split(r'\s*,\s*', exclude.strip())
        self.names = names or []
        self.exclude = exclude or []

    def copy(self, deep=False):
        return type(self)(self.names, self.exclude)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`LeftSelection`

        '''
        data = ensure_element(data)
        return cls(data.text, data.attrib.get('exclude'))

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('left-select',
                            attrib=dict(exclude=','.join(self.exclude) or None),
                            text_content=','.join(self.names))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class RightSelection(object):
    '''
    Join right selection

    Parameters
    ----------
    names : string or list-of-strings
        Specify fields from the left input to the Join window
        to use in output elements.
    exclude : string or list-of-strings, optional
        Specifies a comma-separated list of fields to use
        from the input window.

    Returns
    -------
    :class:`RightSelection`

    '''

    def __init__(self, names, exclude=None):
        if isinstance(names, six.string_types):
            names = re.split(r'\s*,\s*', names.strip())
        if isinstance(exclude, six.string_types):
            exclude = re.split(r'\s*,\s*', exclude.strip())
        self.names = names or []
        self.exclude = exclude or []

    def copy(self, deep=False):
        return type(self)(self.names, self.exclude) 

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`FieldExpression`

        '''
        data = ensure_element(data)
        return cls(data.text, data.attrib.get('exclude'))

    from_xml = from_element

    def to_element(self):
        '''
        Convert object to Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        return xml.new_elem('right-select',
                            attrib=dict(exclude=','.join(self.exclude) or None),
                            text_content=','.join(self.names))

    def to_xml(self, pretty=False):
        '''
        Convert object to XML

        Parameters
        ----------
        pretty : bool, optional
            Should whitespace be added for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)


class JoinWindow(Window, InitializeExpressionFeature):
    '''
    Join window

    Parameters
    ----------
    name : string, optional
        The name of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level
        value of pubsub is manual, true enables publishing and subscribing
        for the window and false disables it.
    description : string, optional
        Description of the window
    output_insert_only : bool, optional
        When true, prevents the window from passing non-insert events to
        other windows.
    collapse_updates : bool, optional
        When true, multiple update blocks are collapsed into a single update block
    pulse_interval : string, optional
        Output a canonical batch of updates at the specified interval
    exp_max_string : int, optional
        Specifies the maximum size of strings that the expression engine
        uses for the window. Default value is 1024.
    index_type : string, optional
        Index type for the window
    pubsub_index_type : string, optional
        Publish/subscribe index type.  Valid values are the same as for the
        `index_type` parameter.
    type : string, optional
        Type of the join
    use_secondary_index : bool, optional
        Use a secondary index
    no_regenerates : bool, optional
        When true, do not regenerate join changes when the dimension
        table changes.
    left_index : string, optional
        Left index type override
    right_index : string, optional
        Right index type override
    conditions : list-of-tuples, optional
        One or more equijoin match pairs.  Each pair should be a two elemnt
        tuple: (left-name, right-name).

    Attributes
    ----------
    expr_initializer : InitializeExpression
        Initialization expression code block
    output : list
        List of FieldPlugins, FieldExpressions, FieldSelections,
        LeftExpressions, RightExpressions, LeftSelections, and RightSelections.

    Returns
    -------
    :class:`JoinWindow`

    '''

    window_type = 'join'

    def __init__(self, name=None, pubsub=None, description=None,
                 output_insert_only=None, collapse_updates=None,
                 pulse_interval=None, exp_max_string=None, index_type=None,
                 pubsub_index_type=None, type='leftouter',
                 use_secondary_index=None, no_regenerates=None,
                 left_index=None, right_index=None, conditions=None):
        Window.__init__(self, **get_args(locals()))
        self.type = type
        self.use_secondary_index = use_secondary_index
        self.no_regenerates = no_regenerates
        self.left_index = left_index
        self.right_index = right_index
        self.conditions = conditions or []
        self.output = []

    def copy(self, deep=False):
        out = Window.copy(self, deep=deep)
        out.type = self.type
        out.use_secondary_index = self.use_secondary_index
        out.no_regenerates = self.no_regenerates
        out.left_index = self.left_index
        out.right_index = self.right_index
        if deep:
            out.conditions = [x.copy(deep=deep) for x in self.conditions]
            out.output = [x.copy(deep=deep) for x in self.output]
        else:
            out.conditions = list(self.conditions)
            out.output = list(self.output)
        return out

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Convert XML / Element to object

        Parameters
        ----------
        data : xml-string or Element
            The element to convert
        session : Session, optional
            The requests.Session object

        Returns
        -------
        :class:`JoinWindow`

        '''
        out = super(JoinWindow, cls).from_element(data, session=session)

        for item in data.findall('./join'):
            for key, value in six.iteritems(item.attrib):
                key = key.replace('-', '_')
                if hasattr(out, key):
                    setattr(out, key, value)
            for cond in item.findall('./conditions/fields'):
                out.conditions.append((cond.attrib['left'], cond.attrib['right']))

        for item in data.findall('./output/field-expr'):
            out.output.append(FieldExpression.from_element(item, session=session))
        for item in data.findall('./output/left-expr'):
            out.output.append(LeftExpression.from_element(item, session=session))
        for item in data.findall('./output/right-expr'):
            out.output.append(LeftExpression.from_element(item, session=session))

        for item in data.findall('./output/field-plug'):
            out.output.append(FieldPlugin.from_element(item, session=session))

        for item in data.findall('./output/field-selection'):
            out.output.append(FieldSelection.from_element(item, session=session))
        for item in data.findall('./output/left-select'):
            out.output.append(LeftSelection.from_element(item, session=session))
        for item in data.findall('./output/right-select'):
            out.output.append(RightSelection.from_element(item, session=session))

        return out

    from_xml = from_element

    def to_element(self, query=None):
        '''
        Convert object to Element

        Parameters
        ----------
        query : string, optional
            Name of the continuous query

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = Window.to_element(self, query=query)

        join = xml.add_elem(out, 'join',
                            attrib=dict(type=self.type,
                                        use_secondary_index=self.use_secondary_index,
                                        no_regenerates=self.no_regenerates,
                                        left_index=self.left_index,
                                        right_index=self.right_index))
        cond = xml.add_elem(join, 'conditions')
        for left, right in self.conditions:
            xml.add_elem(cond, 'fields', attrib=dict(left=left, right=right))

        output = xml.add_elem(out, 'output')
        for item in self.output:
            xml.add_elem(output, item.to_element())

        connectors_to_end(out)

        return out

    def add_condition(self, left, right):
        '''
        Add a join condition

        Parameters
        ----------
        left : string
            Left field for match
        right : string
            Right field for match

        '''
        self.conditions.append((left, right))

    def add_field_expression(self, name, expr, type='double'):
        '''
        Add a field expression

        Parameters
        ----------
        name : string, optional
            Specifies the name of the output field for the join
        expr : string
            Specifies an Expression Engine Language (EEL) expression
            that is assigned to a field.
        type : string, optional
            The data type of the expression result.
            Valid Values: int32, int64, double, string, date, stamp,
            money, blob

        '''
        self.output.append(FieldExpression(name, expr, type=type))

    add_field_expr = add_field_expression

    def add_expression(self, names, type='left'):
        '''
        Add an expression

        Parameters
        ----------
        names : string or list-of-strings
            Specifies fields from the left input or the right input
            to the Join window to use in output elements.
        type : string, optional
            The type of expression.
            Valid values: left or right

        '''
        if type.lower().startswith('r'):
            self.output.append(RightExpression(names))
        else:
            self.output.append(LeftExpression(names))

    add_expr = add_expression

    def add_field_selection(self, name, source):
        '''
        Add a field selection

        Parameters
        ----------
        name : string
            Selected field in the output schema
        source : string
            Takes the following form:l_field_name | rfield_name.,
            where 'l_' indicates that the field comes from the left window
            and 'r_' indicates that the field comes from the right window.

        '''
        self.output.append(FieldSelection(name, source))

    def add_selection(self, selection, type='left', exclude=None):
        '''
        Add a selection

        Parameters
        ----------
        selection : string or list-of-strings
            Specifies fields from the left input or the right input to
            the Join window to use in output elements.
        exclude : string or list-of-strings, optional
            Specifies a comma-separated list of fields to use from
            the input window.
        type : string, optional
            Type of selection.
            Valid values: left or right

        '''
        if type.lower().startswith('r'):
            self.output.append(RightSelection(selection, exclude=exclude))
        else:
            self.output.append(LeftSelection(selection, exclude=exclude))

    def add_field_plugin(self, name, type, plugin, function):
        '''
        Add a field plugin

        Parameters
        ----------
        name : string
            Name of the output field
        type : string
            Type of the output field
        plugin : string
            Full path to .so / .dll
        function : string
            Function to call in the library

        '''
        self.output.append(FieldPlugin(name, type, plugin, function))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestFilterWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('filter_with_expression.xml')

    def test_expression(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('id*:int32','charData:string','int64D:int64',
                                       'doubleD:double','dateD:date',
                                       'timestampD:stamp','moneyD:money'))
        src.add_connector('fs', conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='filter_with_expression.csv',
                          transactional=True,
                          blocksize='3',
                          dateformat='%Y-%m-%d %H:%M:%S')

        filterWin = esp.FilterWindow(name='filter_w')
        filterWin.set_expression('charData=="update\'s result"')
        filterWin.add_connector('fs',
                                conn_name='sub', conn_type='subscribe',
                                type='sub',
                                fstype='csv',
                                fsname='filter.out',
                                snapshot=True,
                                dateformat='%Y-%m-%d %H:%M:%S')

        filterWin2 = esp.FilterWindow(name='filter_w2')
        filterWin2.expression='charData=="insert4"'
        filterWin2.add_connector('fs',
                                 conn_name='sub', conn_type='subscribe',
                                 type='sub',
                                 fstype='csv',
                                 fsname='filter2.out',
                                 snapshot=True,
                                 dateformat='%Y-%m-%d %H:%M:%S')
        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(filterWin)
        cq.add_window(filterWin2)
        src.add_target(filterWin)
        src.add_target(filterWin2)

        self._test_model_file('filter_with_expression.xml', proj)

    @unittest.skip('Need way to load plugin')
    def test_plugin(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',schema=('ID*:int64','symbol:string',
                                                      'quant:int32','price:money'))

        filterWin = esp.FilterWindow(name='filter_w')
        filterWin.set_plugin('libmethod','findTheElement',
                             context_name='libmethod',context_function='get_derived_context')
        filterWin.add_connector('fs',
                                conn_name='fsout', conn_type='subscribe',
                                type='sub',
                                fstype='csv',
                                fsname='output',
                                snapshot=True)
        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(filterWin)
        src.add_target(filterWin)

        self._test_model_file('filter_with_plugin.xml', proj)

    def test_udf(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest')

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win',
                               schema=('id*:int32','charData:string','int64D:int64',
                                       'doubleD:double','dateD:date','moneyD:money'))
        src.add_connector('fs',
                          conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='filter_with_udf.csv',
                          transactional=True,
                          blocksize='4',
                          dateformat='%Y-%m-%d %H:%M:%S')

        filterWin = esp.FilterWindow(name='filter_w')
        filterWin.set_expression_initializer(funcs={
            'udf1:double': 'private real p p = parameter(1); return (p*2)',
            'udf2:string': 'private string p p = parameter(1); return upper(p)'
        })
        filterWin.set_expression('(udf1(doubleD) > 200.0) and match_string(udf2(charData),\'INSERT\')')
        filterWin.add_connector('fs',
                                conn_name='sub', conn_type='subscribe',
                                type='sub',
                                fstype='csv',
                                fsname='filter.out',
                                snapshot=True,
                                dateformat='%Y-%m-%d %H:%M:%S')

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(filterWin)
        src.add_target(filterWin)

        self._test_model_file('filter_with_udf.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
ESP interface

Examples
--------
>>> import esppy
>>> esp = esppy.ESP('http://myesp.com:8777')
>>> esp.server_info
{'analytics-license': True,
 'engine': 'esp',
 'http-admin': 8610,
 'pubsub': 8611,
 'version': '5.2'}

>>> proj = esp.install_project('model.xml')
>>> proj
Project(name='MyProject')

>>> esp.get_running_projects()
{'MyProject': Project(name='MyProject')}

>>> proj.get_windows()
{'w_calculate': CalculationWindow(name='w_calculate',
                                  contquery='contquery',
                                  project='MyProject')),
 'w_data': SourceWindow(name='w_data',
                        contquery='contquery',
                        project='MyProject')),
 'w_request': SourceWindow(name='w_request',
                           contquery='contquery',
                           project='MyProject')}

>>> proj.stop()
>>> esp.get_running_projects()
{}

>>> esp.get_stopped_projects()
{'MyProject': Project(name='MyProject')}

>>> proj.delete()
>>> esp.get_stopped_projects()
{}

>>> esp.shutdown()

'''

from __future__ import print_function, division, absolute_import, unicode_literals

# Configuration
from . import config
from .config import (set_option, get_option, reset_option, describe_option,
                     options, option_context)

from .connection import ESP
from .exceptions import ESPError

__version__ = '1.0.1-dev'



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Modbus Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class ModbusSubscriber(Connector):
    '''
    Subscribe to Modbus events

    Parameters
    ----------
    modbus : string
        The modbus server location. The default port is 502. If your
        server is running on a different port, use host:port notation.
    snapshot : boolean, optional
        If true a snapshot is pulled from the window on startup.

    Returns
    -------
    :class:`ModbusSubscriber`

    '''
    connector_key = dict(cls='modbus', type='subscribe')

    property_defs = dict(
        modbus=prop('modbus', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True)
    )

    def __init__(self, modbus, name=None, is_active=None, snapshot=False):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'modbus', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['modbus'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class ModbusPublisher(Connector):
    '''
    Publish Modbus events

    Parameters
    ----------
    modbus : string
        The modbus server location. The default port is 502. If your
        server is running on a different port, use host:port notation.
    interval : int
        The interval at which the object value data will be pulled
        from the Modbus server (defaults to 10 seconds)

    Returns
    -------
    :class:`ModbusPublisher`

    '''
    connector_key = dict(cls='modbus', type='publish')

    property_defs = dict(
        modbus=prop('modbus', dtype='string', required=True),
        interval=prop('interval', dtype='int')
    )

    def __init__(self, modbus, name=None, is_active=None, interval=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'modbus', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['modbus'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Project '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import os
import re
import requests
import six
import threading
import xml.etree.ElementTree as ET
from six.moves import urllib
from .base import ESPObject, attribute
from .config import get_option
from .contquery import ContinuousQuery
from .exceptions import ESPError
from .mas import MASModule
from .windows import get_window_class
from .utils.rest import get_params
from .utils.data import get_project_data, gen_name
from .utils.events import get_events
from .utils.notebook import scale_svg
from .utils.project import expand_path
from .utils import xml


class ContinuousQueryDict(collections.MutableMapping):
    '''
    Dictionary for holding continuous query objects

    Attributes
    ----------
    project : string
        The name of the project
    session : requests.Session
        The session for the continuous queries

    Parameters
    ----------
    *args : one-or-more arguments, optional
        Positional arguments to MutableMapping
    **kwargs : keyword arguments, optional
        Keyword arguments to MutableMapping

    '''

    def __init__(self, *args, **kwargs):
        collections.MutableMapping.__init__(self, *args, **kwargs)
        self._data = dict()
        self.project = None
        self.session = None

    @property
    def project(self):
        '''
        The project that queries are associated with

        Returns
        -------
        string

        '''
        return self._project

    @project.setter
    def project(self, value):
        self._project = getattr(value, 'name', value)
        for item in self._data.values():
            item.project = self._project

    @property
    def session(self):
        '''
        The session for the queries

        Returns
        -------
        string

        '''
        return self._session

    @session.setter
    def session(self, value):
        self._session = value
        for item in self._data.values():
            item.session = self._session

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        if not isinstance(value, ContinuousQuery):
            raise TypeError('Only ContinuousQuery objects can be '
                            'values in a Project')
        value.name = key
        value.project = self.project
        value.session = self.session
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)


class ConnectorGroup(object):
    '''
    Project connector group

    Attributes
    ----------
    entries : dict
        Collection of connectors

    Parameters
    ----------
    name : string
        The name of the group
    description : string, optional
        A description of the group

    Returns
    -------
    :class:`ConnectorGroup`

    '''

    def __init__(self, name, description=None):
        self.name = name
        self.description = description
        self.entries = {}

    def add_entry(self, connector, state):
        self.entries[connector] = state

    @classmethod
    def from_element(cls, data, session=None):
        if isinstance(data, six.string_types):
            data = xml.from_xml(data)

        out = cls(data.attrib['name'])

        for item in data.findall('./description'):
            out.description = item.text

        for item in data.findall('./connector-entry'):
            out.entries[item.attrib['connector']] = item.attrib['state']

        return out

    from_xml = from_element

    def to_element(self):
        out = xml.new_elem('connector-group', attrib=dict(name=self.name))

        if self.description:
            xml.add_elem(out, 'description', text_content=self.description)

        for key, value in six.iteritems(self.entries):
            xml.add_elem(out, 'connector-entry',
                         attrib=dict(connector=key, state=value))

        return out

    def to_xml(self, pretty=False):
        return xml.to_xml(self.to_element(), pretty=pretty)


class Edge(object):

    def __init__(self, source, target):
        self.source = source
        if isinstance(target, six.string_types):
            self.target = re.split(r'\s*,\s*', target)
        else:
            self.target = list(target)

    @classmethod
    def from_element(cls, data, session=None):
        return cls(data.attrib['source'], data.attrib['target'])

    def to_element(self):
        return xml.new_elem('edge',
                            attrib=dict(source=self.source,
                                        target=','.join(self.target)))


class Project(ESPObject, collections.MutableMapping):
    '''
    ESP Project

    Parameters
    ----------
    name : string
        Name of the project
    n_threads : int, optional
        A positive integer for the number of threads to use from the
        available thread pool
    pubsub : string, optional
        Publish/subscribe mode for the project
    compress_open_patterns : bool, optional
        When True, compress events stored in unresolved pattern instances
    port : int, optional
        The project-level publish/subscribe port for auto or manual mode
    index_type : string, optional
        A default index type for all windows in the project that do not
        explicitly specify an index type
        Valid values: 'rbtree', 'hash', 'ln_hash', 'cl_hash', 'fw_hash', 'empty'
    use_tagged_token : bool, optional
        Specify whether tagged token data flow semantics should be used
        for continuous queries
    use_retention_tracking : bool, optional
        Specify whether to use retention tracking for the project.
        This allows two events per key in output event blocks.
    disk_store_path : string, optional
        Specify the path for on-disk event storage when using hleveldb indices
    heartbeat_interval : int, optional
        Specify how frequently the system sends timer heartbeats to the project.
        The default value is 1 second. This heartbeat drives wall-clock-based
        retention and pattern time outs. When a model has an exceptionally large
        number of unresolved pattern instances, increasing the value of the
        heartbeat interval can optimize how the model runs.
    restore : string, optional
        Specify a directory from which to restore the project from a
        previously-performed persist operation.
    description : string, optional
        Description of the project
    metadata : dict, optional
        Metadata associated with the project

    Attributes
    ----------
    queries : dict
        Dictionary of the queries
    metadata : dict
        Metadata associated with the project
    url : string
        The URL of the project

    Notes
    -----
    All parameters to the constructor are also available as attributes.

    Returns
    -------
    :class:`Project`

    '''

    default_query = 'contquery'

    n_threads = attribute('threads', dtype='int')
    pubsub = attribute('pubsub', dtype='string', values=['none', 'auto', 'manual'])
    port = attribute('port', dtype='int')
    index_type = attribute('index', dtype='string',
                           values={'rbtree': 'pi_RBTREE', 'hash': 'pi_HASH',
                                   'ln_hash': 'pi_LN_HASH', 'cl_hash': 'pi_CL_HASH',
                                   'fw_hash': 'pi_FW_HASH', 'empty': 'pi_EMPTY'})
    name = attribute('name', dtype='string')
    use_tagged_token = attribute('use-tagged-token', dtype='bool')
    use_retention_tracking = attribute('retention-tracking', dtype='bool')
    restore = attribute('restore', dtype='string')
    disk_store_path = attribute('disk-store-path', dtype='string')
    heartbeat_interval = attribute('heartbeat-interval', dtype='int')
    compress_open_patterns = attribute('compress-open-patterns', dtype='bool')

    def __init__(self, name=None, n_threads=1, pubsub='auto',
                 compress_open_patterns=None, port=None, index_type=None,
                 use_tagged_token=None, use_retention_tracking=None,
                 disk_store_path=None, heartbeat_interval=None, restore=None,
                 description=None, sas_log_location=None, sas_connection_key=None,
                 sas_command=None, metadata=None):
        self.queries = ContinuousQueryDict()
        ESPObject.__init__(self, attrs=locals())
        self.name = name or gen_name(prefix='p_')
        self.description = description
        self.sas_log_location = sas_log_location
        self.sas_connection_key = sas_connection_key
        self.sas_command = sas_command
        self.mas_modules = []
        self.connector_groups = {}
        self.edges = []
        self.metadata = dict(metadata or {})

    @property
    def session(self):
        '''
        The requests.Session object for the project

        Returns
        -------
        string

        '''
        return ESPObject.session.fget(self)

    @session.setter
    def session(self, value):
        ESPObject.session.fset(self, value)
        self.queries.session = value

    @property
    def name(self):
        '''
        The name of the project

        Returns
        -------
        string

        '''
        return self._name

    @name.setter
    def name(self, value):
        '''
        Set the name of the project and propagate to sub-objects

        Parameters
        ----------
        value : string
            The name of the project

        '''
        self._name = value
        self.queries.project = self.name

    @property
    def windows(self):
        '''
        Return collection of windows from the default query

        Returns
        -------
        dict

        '''
        if self.default_query not in self.queries:
            self.queries[self.default_query] = ContinuousQuery()
        return self.queries[self.default_query].windows

    @property
    def trace(self):
        '''
        Set the trace property on the default continuous query

        Returns
        -------
        string

        '''
        if self.default_query not in self.queries:
            self.queries[self.default_query] = ContinuousQuery()
        return self.queries[self.default_query].trace

    @trace.setter
    def trace(self, value):
        '''
        Set the trace property on the default continuous query

        Parameters
        ----------
        value : string
            The names of the windows to trace

        '''
        if self.default_query not in self.queries:
            self.queries[self.default_query] = ContinuousQuery()
        self.queries[self.default_query].trace = value

    def sync(self, overwrite=True, start=True, start_connectors=True):
        '''
        Sync the project definition to the server

        Notes
        -----
        This will overwrite the project definition in the server.

        Parameters
        ----------
        overwrite : bool, optional
            Should an existing project with the same name be overwritten?
        start : bool, optional
            Should the project be started?
        start_connectors : bool, optional
            Should the connectors be started?

        '''
        self._put(params=get_params(overwrite=overwrite,
                                    connectors=start_connectors,
                                    start=start),
                  data=get_project_data(self))

    def copy(self, deep=False):
        '''
        Return a copy of the object

        Parameters
        ----------
        deep : bool, optional
            Should sub-objects be copied as well?

        Returns
        -------
        :class:`Project`

        '''
        out = type(self)()

        out.session = self.session

        for key, value in self._get_attributes(use_xml_values=False).items():
            setattr(out, key, value)

        if deep:
            out.queries = dict([(k, v.copy(deep=True))
                                for k, v in self.queries.items()])
        else:
            out.queries.update(self.queries)

        return out

    def __copy__(self):
        return self.copy(deep=False)

    def __deepcopy__(self, memo):
        return self.copy(deep=True)

    @property
    def fullname(self):
        return self.name

    @property
    def url(self):
        '''
        URL of project

        Returns
        -------
        string

        '''
        return urllib.parse.urljoin(self.base_url, 'projects/%s/' % self.name)

    @classmethod
    def from_element(cls, data, session=None):
        '''
        Create project from XML definition

        Parameters
        ----------
        data : xml-string or ElementTree.Element
            XML project definition
        session : requests.Session, optional
            Session that the project is associated with

        Returns
        -------
        :class:`Project`

        '''
        out = cls()
        out.session = session

        if isinstance(data, six.string_types):
            data = xml.from_xml(get_project_data(data))

        if data.tag != 'project':
            data = data.find('.//project')

        if data is None:
            raise ValueError('No project found in input.')

        out._set_attributes(data.attrib)

        for desc in data.findall('./description'):
            out.description = desc.text

        for item in data.findall('./ds-initialize'):
            out.sas_log_location = item.attrib.get('sas-log-location')
            out.sas_connection_key = item.attrib.get('sas-connection-key')
            out.sas_command = item.attrib.get('sas-command')

        for item in data.findall('./mas-modules/mas-module'):
            out.mas_modules.append(MASModule.from_xml(item, session=session))

        for item in data.findall('./project-connectors/connector-groups/connector-group'):
            grp = ConnectorGroup.from_element(item, session=session)
            out.connector_groups[grp.name] = grp

        for item in data.findall('./project-connectors/edges/edge'):
            out.edges.append(Edge.from_element(item, session=session))

        for contquery in data.findall('.//contquery'):
            query = ContinuousQuery.from_xml(contquery, project=out,
                                             session=session)
            out.queries[query.name] = query

        for item in data.findall('./metadata/meta'):
            out.metadata[item.attrib['id']] = item.text

        return out

    from_xml = from_element

    def _load_metadata(self):
        data = self._get(urllib.parse.urljoin(self.base_url,
                                              'projectMetadata/%s' % self.name))
        for item in data.findall('./project/metadata/meta'):
            self.metadata[item.attrib['id']] = item.text
        for cq in data.findall('./project/contquery'):
            name = cq.attrib['id']
            if name in self.queries:
                for item in cq.findall('./metadata/meta'):
                    self.queries[name].metadata[item.attrib['id']] = item.text

    def _persist_metadata(self):
        if self.metadata:
            self._set_metadata(self.metadata)
        for cq in six.itervalues(self.queries):
            if cq.metadata:
                cq._set_metadata(cq.metadata)

    def _clear_metadata(self):
        self.metadata.clear()
        for cq in six.itervalues(self.queries):
            cq._clear_metadata()

    def _get_metadata(self):
        out = {}
        for item in self._get(urllib.parse.urljoin(self.base_url,
                                                   'projectMetadata/%s' % self.name)
                             ).findall('./project/metadata/meta'):
            out[item.attrib['id']] = item.text
        return out

    def _set_metadata(self, data):
        for key, value in six.iteritems(data):
            self._put(urllib.parse.urljoin(self.base_url,
                                           'projectMetadata/%s/%s' % (self.name, key)),
                      data='%s' % value)

    def _del_metadata(self, *data):
        for key in data:
            self._delete(urllib.parse.urljoin(self.base_url,
                                              'projectMetadata/%s/%s' % (self.name, key)))

    def to_element(self):
        '''
        Export project definition to an ElementTree.Element

        Returns
        -------
        :class:`ElementTree.Element`

        '''
        out = xml.new_elem('engine')
        projects = xml.add_elem(out, 'projects')

        proj = xml.add_elem(projects, 'project', xml.get_attrs(self, extra='name'))

        if self.description:
            xml.add_elem(proj, 'description', text_content=self.description)

        if self.metadata:
            metadata = xml.add_elem(proj, 'metadata')
            for key, value in sorted(six.iteritems(self.metadata)):
                xml.add_elem(metadata, 'meta', attrib=dict(id=key),
                             text_content=value)

        if self.sas_log_location or self.sas_connection_key or self.sas_command:
            xml.add_elem(proj, 'ds-initialize',
                         attrib=dict(sas_log_location=self.sas_log_location,
                                     sas_connection_key=self.sas_connection_key,
                                     sas_command=self.sas_command))

        if self.mas_modules:
            mods = xml.add_elem(proj, 'mas-modules')
            for item in self.mas_modules:
                xml.add_elem(mods, item.to_element())

        queries = xml.add_elem(proj, 'contqueries')

        for name, query in sorted(six.iteritems(self.queries)):
            xml.add_elem(queries, query.to_element())

        if self.connector_groups or self.edges:
            conns = xml.add_elem(proj, 'project-connectors')

            if self.connector_groups:
                grps = xml.add_elem(conns, 'connector-groups')
                for name, value in sorted(six.iteritems(self.connector_groups)):
                    xml.add_elem(grps, value.to_element())

            if self.edges:
                edges = xml.add_elem(conns, 'edges')
                for item in self.edges:
                    xml.add_elem(edges, item.to_element())

        return out

    def to_xml(self, pretty=False):
        '''
        Export project definition to XML

        Parameters
        ----------
        pretty : bool, optional
            Should the XML include whitespace for readability?

        Returns
        -------
        string

        '''
        return xml.to_xml(self.to_element(), pretty=pretty)

    def save_xml(self, dest, mode='w', pretty=True, **kwargs):
        '''
        Save the project XML to a file

        Parameters
        ----------
        dest : string or file-like
            The destination of the XML content
        mode : string, optional
            The write mode for the output file (only used if `dest` is a string)
        pretty : boolean, optional
            Should the XML include whitespace for readability?

        '''
        if isinstance(dest, six.string_types):
            with open(dest, mode=mode, **kwargs) as output:
                output.write(self.to_xml(pretty=pretty))
        else:
            dest.write(self.to_xml(pretty=pretty))

    def to_graph(self, graph=None, schema=False):
        '''
        Export project definition to graphviz.Digraph

        Parameters
        ----------
        graph : graphviz.Digraph, optional
            The parent graph to add to
        schema : bool, optional
            Should window schemas be included?

        Returns
        -------
        :class:`graphviz.Digraph`

        '''
        try:
            import graphviz as gv
        except ImportError:
            raise ImportError('The graphviz module is required for exporting to graphs.')

        if graph is None:
            graph = gv.Digraph(format='svg')
            graph.attr('node', shape='rect', fontname='helvetica')
            graph.attr('graph', rankdir='LR', center='false')
            graph.attr('edge', fontname='times-italic')

        if self.queries:
            pgraph = gv.Digraph(format='svg',
                                name='cluster_%s' % self.fullname.replace('.', '_'))
            pgraph.attr('graph', fontname='helvetica')
            pgraph.attr('edge', fontname='times-italic')
            pgraph.attr(label=self.name, labeljust='l',
                        style='filled,bold,rounded', color='#c0c0c0',
                        fillcolor='#dadada', fontcolor='black')

            if list(self.queries.keys()) == [self.default_query]:
                for wkey, win in sorted(self.windows.items()):
                    pgraph.subgraph(win.to_graph(schema=schema))
                    for target in win.targets:
                        graph.edge(win.fullname, self.windows[target.name].fullname,
                                   label=target.role or '')
            else:
                for qkey, query in sorted(self.queries.items()):
                    pgraph.subgraph(query.to_graph(schema=schema))

            graph.subgraph(pgraph)

        else:
            graph.node(self.fullname, label=self.name, labeljust='l',
                       style='filled,bold,rounded', color='#c0c0c0',
                       fillcolor='#dadada', fontcolor='black')

        return graph

    def _repr_svg_(self):
        try:
            return scale_svg(self.to_graph()._repr_svg_())
        except ImportError:
            raise AttributeError('_repr_svg_')

    def __str__(self):
        return '%s(name=%s)' % (type(self).__name__, repr(self.name))

    def __repr__(self):
        return str(self)

    def start(self):
        ''' Start the project '''
        self._put('state', params=dict(value='running'))

    def stop(self):
        ''' Stop the project '''
        self._put('state', params=dict(value='stopped'))

    def start_connectors(self):
        ''' Start the connectors for the project '''
        self._put('state', params=dict(value='connectorsStarted'))

    def add_connectors(self, group_name, connectors, group_description=None):
        '''
        Add connectors to a group

        Parameters
        ----------
        group_name : string
            The name of the connector group
        connectors : dict
            Key / value pairs of connector names and states
            ('finished', 'running', 'stopped')
        group_description : string, optional
            A description of the group

        Notes
        -----
        If the connector group does not exist yet, one is created.
        If the connector group does exist, the connectors will be
        added to it.

        '''
        if group_name not in self.connector_groups:
            self.connector_groups[group_name] = ConnectorGroup(group_name,
                                                    description=group_description)

        group = self.connector_groups[group_name]

        for key, value in six.iteritems(connectors):
            group.add_entry(key, value)

    def add_edge(self, source, targets):
        '''
        Add connector edges

        Parameters
        ----------
        source : string
           Name of the source
        targets : string or list-of-strings
           If a string, then it is a single target.  If a list, then
           it is a list of target names.

        '''
        if isinstance(targets, six.string_types):
            targets = [targets]
        self.edges.append(Edge(source, targets)) 

    def save(self, path=None):
        '''
        Save the project

        Parameters
        ----------
        path : string
            Directory to save the project in

        '''
        self._put('state', params=get_params(value='persisted', path=path))

    def restore(self, path=None):
        '''
        Restore the project

        Parameters
        ----------
        path : string
            Directory where project was saved

        '''
        self._put('state', params=get_params(value='restored', path=path))

    def update(self, project):
        '''
        Update project definition

        Parameters
        ----------
        project : string
            Location of the project data

        '''
        self._put('state', params=get_params(value='modified'),
                  data=get_project_data(project))

    def delete(self):
        ''' Delete the project '''
        self._delete()

    def create_mas_module(self, language, module, func_names, mas_store=None,
                          mas_store_version=None, description=None,
                          code_file=None, code=None):
        '''
        Create a MAS module object

        Parameters
        ----------
        language : string
            The name of the programming language
        module : string, optional
            Name of the MAS module
        func_names : string or list-of-strings, optional
            The function names exported by the module

        Returns
        -------
        :class:`MASModule`

        '''
        out = MASModule(language, module, func_names, mas_store=mas_store,
                        mas_store_version=mas_store_version,
                        description=description, code_file=code_file,
                        code=code)
        out.project = self.name
        out.session = self.session
        return out

    def get_mas_modules(self, expandcode=False):
        '''
        Retrieve all MAS modules

        Parameters
        ----------
        expandcode : bool, optional
            If True, return code-file elements as code elements containing
            the contents of the file.

        Returns
        -------
        list of :class:`MASModule`s

        '''
        res = self._get(urllib.parse.urljoin(self.base_url,
                                             'masModules/%s' % self.name),
                        params=get_params(expandcode=expandcode))
        modules = []
        for elem in res.findall('./project'):
            for mod in elem.findall('./mas-module'):
                mod = MASModule.from_xml(mod, session=self.session)
                mod.project = self.name
                modules.append(mod)
        return modules

    def get_mas_module(self, name, expandcode=False):
        '''
        Retrieve specified MAS modules

        Parameters
        ----------
        name : string
            Name of the MAS module
        expandcode : bool, optional
            If True, return code-file elements as code elements containing
            the contents of the file.

        Returns
        -------
        :class:`MASModule`

        '''
        res = self._get(urllib.parse.urljoin(self.base_url,
                                             'masModules/%s/%s' %
                                             (self.name, name)),
                        params=get_params(expandcode=expandcode))
        mod = MASModule.from_xml(res, session=self.session)
        mod.project = self.name
        return mod

    def replace_mas_module(self, name, module):
        '''
        Replace the specified MAS module

        Parameters
        ----------
        name : string
            Name of the MAS module
        module : string or MASModule
            MAS module configuration

        '''
        self._put(urllib.parse.urljoin(self.base_url,
                                       'masModules/%s/%s' % (self.name, name)),
                  data=get_project_data(module))

    def get_stats(self, interval=None, min_cpu=None, limit=20):
        '''
        Retrieve project statistics subscriber

        Parameters
        ----------
        interval : int, optional
            The interval in seconds between updates
        min_cpu : int, optional
            The minimum CPU value you want included
        limit : int, optional
            The maximum number of rows to retain in the DataFrame

        Returns
        -------
        :class:`ProjectStats`

        '''
        from .connection import ProjectStats
        proj_stats = ProjectStats(self, interval=interval, min_cpu=min_cpu,
                                  limit=limit)
        proj_stats.start()
        return proj_stats

    def validate(self):
        '''
        Validate the project definition

        Returns
        -------
        True
            If project is valid
        False
            If project is not valid

        '''
        try:
            res = self._post(urllib.parse.urljoin(self.base_url,
                                                  'projectValidationResults'),
                             data=self.to_xml())
            if res.tag != 'schema-validation-success':
                return False
        except ESPError:
            return False
        return True

    def add_window(self, window, contquery=None):
        '''
        Add a window to the project

        Parameters
        ----------
        window : Window
            The Window object to add
        contquery : string or ContinuousQuery, optional
            The name of the continuous query the window belongs to.
            If not specified, the default continuous query will
            be used.

        Returns
        -------
        :class:`Window`

        '''
        if contquery is None:
            contquery = self.default_query
        else:
            contquery = getattr(contquery, 'name', contquery)

        if contquery not in self.queries:
            self.queries[contquery] = ContinuousQuery()

        self.queries[contquery].add_window(window)

        return window

    def add_query(self, contquery):
        '''
        Add a continuous query object

        Parameters
        ----------
        contquery : string or ContinuousQuery
            The name of the continuous query to create, or a
            ContinuousQuery object to add to the project

        Returns
        -------
        :class:`ContinuousQuery`

        '''
        if isinstance(contquery, six.string_types):
            self.queries[contquery] = ContinuousQuery()
            return self.queries[contquery]
        else:
            if not contquery.name:
                contquery.name = gen_name(prefix='cq_')
            self.queries[contquery.name] = contquery
            return self.queries[contquery.name]

    add_continuous_query = add_query

    add_contquery = add_query

    def get_windows(self, path='*.*', type=None, filter=None):
        '''
        Retrieve windows from the server

        Parameters
        ----------
        path : string, optional
            '.' delimited path to windows to retrieve.  '*' can be used as
            a wildcard for any component.  Multiple component names can
            be delimited by '|'.
        type : string or list-of-strings, optional
            Types of windows you want to retrieve
        filter : string or list-of-strings, optional
            Function filter indicating which windows to retrieve

        Notes
        -----
        This method retrieves window definitions from the server, not
        from the collection of windows held locally in this projects
        variables.

        Returns
        -------
        list of :class:`Window`

        '''
        path = [None, None] + expand_path(path)
        res = self._get(urllib.parse.urljoin(self.base_url, 'windowXml'),
                        params=get_params(project=self.name,
                                          contquery=path[-2],
                                          name=path[-1], type=type,
                                          filter=filter))
        windows = {}
        for item in res.findall('./*'):
            try:
                wcls = get_window_class(item.tag)
            except KeyError:
                raise TypeError('Unknown window type: %s' % item.tag)
            window = wcls.from_xml(item, session=self.session)
            windows[window.name] = window
        return windows

    def get_window(self, path):
        '''
        Retrieve specified window

        Parameters
        ----------
        path : string
            '.' delimited path to the window including continuous query
            name and window name.

        Notes
        -----
        This method retrieves window definitions from the server, not
        from the collection of windows held locally in this projects
        variables.

        Returns
        -------
        :class:`Window`

        '''
        out = self.get_windows(path)
        if out and len(out) == 1:
            return out[list(out.keys())[0]]
        raise KeyError("No window with the path '%s'." % path)

#
# MutableMapping methods
#

    def __getitem__(self, key):
        return self.queries[key]

    def __setitem__(self, key, value):
        value.project = self.name
        self.queries[key] = value

    def __delitem__(self, key):
        del self.queries[key]

    def __iter__(self):
        return iter(self.queries)

    def __len__(self):
        return len(self.queries)

    def __contains__(self, value):
        return value in self.queries



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP SMTP Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class SMTPSubscriber(Connector):
    '''
    Subscribe to Simple Mail Transfer Protocol (SMTP) events

    Parameters
    ----------
    smtpserver : string
       Specifies the SMTP server host name or IP address
    sourceaddress : string
        Specifies the e-mail address to be used in the “from” field of
        the e-mail.
    destaddress : string
        Specifies the e-mail address to which to send the e-mail message
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable. The default value is disabled.
    emailperevent : boolean, optional
        Specifies true or false. The default is false. If false, each
        e-mail body contains a full event block. If true, each mail
        body contains a single event.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks
        received by a subscriber that were introduced by a window
        retention policy.
    configfilesection : string, optional
        Specifies the name of the section in the connoctor config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields
        in CSV events. The default behavior is these fields are
        interpreted as an integer number of seconds (ESP_DATETIME)
        or microseconds (ESP_TIMESTAMP) since epoch.

    Returns
    -------
    :class:`SMTPSubscriber`

    '''
    connector_key = dict(cls='smtp', type='subscribe')

    property_defs = dict(
        smtpserver=prop('smtpserver', dtype='string', required=True),
        sourceaddress=prop('sourceaddress', dtype='string', required=True),
        destaddress=prop('destaddress', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        collapse=prop('collapse', dtype='string'),
        emailperevent=prop('emailperevent', dtype='boolean'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        configfilesection=prop('configfilesection', dtype='string'),
        dateformat=prop('dateformat', dtype='string')
    )

    def __init__(self, smtpserver=None, sourceaddress=None, destaddress=None,
                 name=None, is_active=None, snapshot=None,
                 collapse=None, emailperevent=None,
                 rmretdel=None, configfilesection=None, dateformat=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'smtp', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['smtpserver',
                                                   'sourceaddress',
                                                   'destaddress'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)
